from dataclasses import dataclass
from functools import cached_property

from lxml import etree
from lxml.html import HtmlElement, HTMLParser
from lxml.html.clean import Cleaner

from plainhtml import utils
from plainhtml.utils import DOUBLE_NEWLINE_TAGS, NEWLINE_TAGS, Context, DoubleNewline, Newline

DOUBLE_NEWLINE = DoubleNewline()
NEWLINE = Newline()


@dataclass
class Parser:
    html: str

    @cached_property
    def cleaner(self) -> Cleaner:
        """[summary]

        Returns
        -------
        Cleaner
            [description]
        """
        return Cleaner(
            scripts=True,
            javascript=True,
            comments=True,
            style=True,
            inline_style=True,
            links=True,
            meta=True,
            page_structure=True,
            processing_instructions=True,
            embedded=True,
            frames=True,
            forms=True,
            annoying_tags=True,
            remove_unknown_tags=True,
        )

    @cached_property
    def tree(self) -> HtmlElement:
        """[summary]

        Returns
        -------
        HtmlElement
            [description]
        """
        body = self.html.strip().replace("\x00", "").encode("utf-8")
        parser = HTMLParser(recover=True, encoding="utf-8")
        tree = etree.fromstring(body, parser=parser)
        return self.cleaner.clean_html(tree)

    def extract(self) -> str:
        """[summary]

        Returns
        -------
        str
            [description]
        """
        self.chunks = []
        self.context = Context(DOUBLE_NEWLINE)
        self.traverse(self.tree)
        return "".join(self.chunks).strip()

    def traverse(self, tree: HtmlElement):
        """[summary]

        Parameters
        ----------
        tree : HtmlElement
            [description]
        """
        self.add_newlines(tree.tag)
        self.add_text(tree.text)
        for child in tree:
            self.traverse(child)
        self.add_newlines(tree.tag)

    def add_newlines(self, tag: str):
        """[summary]

        Parameters
        ----------
        tag : str
            [description]
        """
        if not self.context.previous is DOUBLE_NEWLINE:
            if tag in DOUBLE_NEWLINE_TAGS:
                self.chunks.append("\n" if self.context.previous is NEWLINE else "\n\n")
                self.context.previous = DOUBLE_NEWLINE
            elif tag in NEWLINE_TAGS:
                if not self.context.previous is NEWLINE:
                    self.chunks.append("\n")
                self.context.previous = NEWLINE

    def add_text(self, text: str):
        """[summary]

        Parameters
        ----------
        text : str
            [description]
        """
        if (text := utils.normalize_whitespace(text)) :
            space = self.space_between(text)
            self.chunks.extend([space, text])
            self.context.previous = text

    def space_between(self, text: str) -> str:
        """[summary]

        Parameters
        ----------
        text : str
            [description]

        Returns
        -------
        str
            [description]
        """
        if text:
            return " " if self.should_add_space(text) else ""
        else:
            return " "

    def should_add_space(self, text: str) -> bool:
        """[summary]

        Parameters
        ----------
        text : str
            [description]

        Returns
        -------
        bool
            [description]
        """
        # TODO: find other name
        if self.context.previous in {NEWLINE, DOUBLE_NEWLINE}:
            return False
        if not utils.has_trailing_whitespace(self.context.previous):
            if utils.has_punctuation_after(text) or utils.has_open_bracket_before(
                self.context.previous
            ):
                return False
        return True



from plainhtml.parser import Parser


def extract(html: str) -> str:
    """[summary]

    Parameters
    ----------
    html : str
        [description]

    Returns
    -------
    str
        [description]
    """
    parser = Parser(html)
    return parser.extract()



import re
from dataclasses import dataclass
from typing import Optional, Union

NEWLINE_TAGS = {
    "article",
    "aside",
    "br",
    "dd",
    "details",
    "div",
    "dt",
    "fieldset",
    "figcaption",
    "footer",
    "form",
    "header",
    "hr",
    "legend",
    "li",
    "main",
    "nav",
    "table",
    "tr",
}
DOUBLE_NEWLINE_TAGS = {
    "blockquote",
    "dl",
    "figure",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "ol",
    "p",
    "pre",
    "title",
    "ul",
}
WHITESPACE = re.compile(r"\s+")
TRAILING_WHITESPACE = re.compile(r"\s$")
PUNCTUATION_AFTER = re.compile(r'^[,:;.!?")]')
OPEN_BRACKET_BEFORE = re.compile(r"\($")


def has_trailing_whitespace(text: str) -> bool:
    """[summary]

    Parameters
    ----------
    text : str
        [description]

    Returns
    -------
    bool
        [description]
    """
    if TRAILING_WHITESPACE.search(text):
        return True
    else:
        return False


def has_punctuation_after(text: str) -> bool:
    """[summary]

    Parameters
    ----------
    text : str
        [description]

    Returns
    -------
    bool
        [description]
    """
    if PUNCTUATION_AFTER.search(text):
        return True
    else:
        return False


def has_open_bracket_before(text: str) -> bool:
    """[summary]

    Parameters
    ----------
    text : str
        [description]

    Returns
    -------
    bool
        [description]
    """
    if OPEN_BRACKET_BEFORE.search(text):
        return True
    else:
        return False


def normalize_whitespace(text: Optional[str]) -> str:
    """[summary]

    Parameters
    ----------
    text : Optional[str]
        [description]

    Returns
    -------
    str
        [description]
    """
    if text:
        return WHITESPACE.sub(" ", text.strip())
    else:
        return ""


class DoubleNewline:
    pass


class Newline:
    pass


@dataclass
class Context:
    previous: Union[DoubleNewline, Newline]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['plainhtml']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.6.2,<5.0.0']

setup_kwargs = {
    'name': 'plainhtml',
    'version': '0.1.1',
    'description': 'Extract plain text from HTML',
    'long_description': '# Extract plain text from HTML\n\n## Installation\n\n```\n$ pip install plainhtml\n```\n\n## Example\n\n```python\n>>> import plainhtml\n>>> html = "<html><body><p>foo</p><p>bar</p></body></html>"\n>>> plainhtml.extract(html)\n\'foo\\n\\nbar\'\n```\n',
    'author': 'Severin Simmler',
    'author_email': 's.simmler@snapaddy.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
